// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#pragma once

#include <iostream>	 // IWYU pragma: keep
#include <sstream>	 // IWYU pragma: keep
#include <string>		 // for string, operator<<

#include "emper-config.h"	 // IWYU pragma: keep

#ifdef EMPER_LOG_OFF

#define DBG(x)
#define WDBG(x)
#define LOGD(x)
#define LOGDD(x)

#else

// If we apply clang-format to the following region, then clang-format
// will create multi-line macros. However clang-tidy's NOLINTNEXTLINE
// would then not work. :(

// clang-format off

// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define DBG(x) do { std::stringstream sst; sst << x << std::endl; emper_log("", sst.str()); } while (false)
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define WDBG(x)	do { std::stringstream sst;	sst << x; emper_log("", sst.str()); } while (false)
// To avoid "error: there are no arguments to ‘logD’ that depend on a
// template parameter, so a declaration of ‘logD’ must be available"
// we use "this->logD()" instead of simply "logD()" below.
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGD(x)	do { std::stringstream sst;	sst << x; this->logD(sst.str()); } while (false)
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGDD(x) do { std::stringstream sst; sst << x; this->logDD(sst.str()); } while (false)

#endif

// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGI(x)	do { std::cerr << "Info: " << x << std::endl; } while (false)
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGW(x)	do { std::cerr << "Warning: " << x << std::endl; } while (false)
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGE(x)	do {std::cerr << "Error: " << x << std::endl; } while (false)

// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define ABORT(x) do { std::stringstream sst; sst << x; logI(sst.str()); abort(); } while (false)

// clang-format on

#define IGNORE_UNUSED_FUNCTION \
	_Pragma("GCC diagnostic push") _Pragma("GCC diagnostic ignored \"-Wunused-function\"")

#define POP_DIAGNOSTIC _Pragma("GCC diagnostic pop")

enum class LogSubsystem {
	PS,
	F,
	C,
	CM,
	DISP,
	SCHED,
	RUNTI,
	U_B_MPSC_Q,
	IO,
};

enum LogLevel {
	OFF,
	Error,
	Warning,
	Info,
	Debug,
	FineDebug,
	FinerDebug,
	FinestDebug,
	ALL,
};

// global log level
namespace emper {
extern enum LogLevel log_level;
}

void emper_add_timestamp_to(std::ostringstream& logMessage);

void emper_log(const std::string& prefix, const std::string& message);

template <LogSubsystem logSubsystem>
class Logger {
 private:
	static constexpr auto getLevelFor(LogSubsystem system) -> LogLevel {
		switch (system) {
			case LogSubsystem::PS:
			case LogSubsystem::F:
			case LogSubsystem::C:
			case LogSubsystem::CM:
			case LogSubsystem::DISP:
			case LogSubsystem::SCHED:
			case LogSubsystem::RUNTI:
			case LogSubsystem::U_B_MPSC_Q:
			case LogSubsystem::IO:
			default:
				return ALL;
		}
	}

	static constexpr auto getTagFor(LogSubsystem system) -> char const* {
		switch (system) {
			case LogSubsystem::PS:
				return "PS   ";
			case LogSubsystem::F:
				return "F    ";
			case LogSubsystem::C:
				return "C    ";
			case LogSubsystem::CM:
				return "CM   ";
			case LogSubsystem::DISP:
				return "DISP ";
			case LogSubsystem::SCHED:
				return "SCHED";
			case LogSubsystem::RUNTI:
				return "RUNTI";
			case LogSubsystem::U_B_MPSC_Q:
				return "UBSCQ";
			case LogSubsystem::IO:
				return "IO   ";
			default:
				return "UNKNOWN SUBSYSTEM (Add it *now*)";
		}
	}

	static constexpr auto shouldPrefixThis(LogSubsystem system) -> bool {
		switch (system) {
			case LogSubsystem::RUNTI:
			case LogSubsystem::SCHED:
				return false;
			default:
				return true;
		}
	}

 protected:
	inline void log(LogLevel level, const std::string& string) const {
#ifdef EMPER_LOG_OFF
		return;
#endif
		// check global level
		if (level > emper::log_level) return;
		// check subsystem level
		if (level > getLevelFor(logSubsystem)) return;

		std::string subSystemTag = getTagFor(logSubsystem);

		std::ostringstream sst;
		sst << subSystemTag;
		if (shouldPrefixThis(logSubsystem)) {
			sst << " " << this;
		}

		emper_log(sst.str(), string);
	}

	inline void logE(const std::string& string) const { log(Error, string); }

	inline void logI(const std::string& string) const { log(Info, string); }

	inline void logD(const std::string& string) const { log(Debug, string); }

	inline void logDD(const std::string& string) const { log(FineDebug, string); }
};
