// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus, Florian Fischer
#include "Runtime.hpp"

#include <pthread.h>	// for pthread_t, pthread_attr_init

#include <cerrno>	 // for errno
// Non portable.
#include <sched.h>				// for cpu_set_t, CPU_SET, CPU_ZERO
#include <sys/sysinfo.h>	// for get_nprocs

#include <cstdlib>	// for rand, srand, abort
#include <cstring>
#include <memory>	 // for __shared_ptr_access, shared_ptr
#include <string>	 // for string

#include "Common.hpp"					 // for DIE_MSG_ERRNO, DIE, DIE_MSG
#include "ContextManager.hpp"	 // for ContextManager
#include "Debug.hpp"					 // for DBG, ABORT, LOGD, LOGE
#include "Emper.hpp"
#include "Fiber.hpp"	// for Fiber
#include "NextFiberResult.hpp"
#include "RuntimeStrategy.hpp"	// for RuntimeStrategy
#include "RuntimeStrategyFactory.hpp"
#include "RuntimeStrategyStats.hpp"	 // for RuntimeStrategyStats
#include "emper-config.h"						 // IWYU pragma: keep
#include "io/IoContext.hpp"					 // for IoContext
#include "io/Stats.hpp"							 // for emper::io::Stats
#include "lib/DebugUtil.hpp"

#ifdef EMPER_DEFAULT_SCHEDULING_STRATEGY_WORK_STEALING
#include "strategies/ws/WsStrategyFactory.hpp"
#elif defined EMPER_DEFAULT_SCHEDULING_STRATEGY_LOCALITY_AWARE_WORK_STEALING
#include "strategies/laws/LawsStrategyFactory.hpp"
#else
#error "Unknown default scheduling strategy"
#endif

#ifdef EMPER_LIBURCU
#include <urcu.h>	 // for rcu_register_thread
#endif

#ifndef EMPER_LOG_OFF
#include <syscall.h>	// for SYS_gettid
#include <unistd.h>		// for syscall

#include <ostream>	// for operator<<, basic_ostream<>:...
#endif

std::mutex Runtime::currentRuntimeMutex;
Runtime* Runtime::currentRuntime;

RuntimeStrategyFactory& Runtime::DEFAULT_STRATEGY =
#ifdef EMPER_DEFAULT_SCHEDULING_STRATEGY_WORK_STEALING
		WsStrategyFactory::INSTANCE
#elif defined EMPER_DEFAULT_SCHEDULING_STRATEGY_LOCALITY_AWARE_WORK_STEALING
		LawsStrategyFactory::INSTANCE
#else
#error "Unknown default scheduling strategy"
#endif
		;

using emper::io::IoContext;

Runtime::Runtime(workerid_t workerCount, RuntimeStrategyFactory& strategyFactory, unsigned int seed)
		: workerCount(workerCount),
			workerLatch(workerCount),
			strategy(strategyFactory.constructRuntimeStrategy(*this)),
			scheduler(strategy->getScheduler()),
			dispatcher(strategy->getDispatcher()),
			contextManager(*(new ContextManager(*this))),
			threads(new pthread_t[workerCount]),
			workers(new Worker*[workerCount]),
			randomEngine(seed),
			sleepingWorkers(0),
			skipSleep(false) {
	const int nprocs = get_nprocs();

	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		if (currentRuntime != nullptr) {
			DIE_MSG("There is already a runtime active");
		}
		currentRuntime = this;
	}

	// initialize the global and all worker IoContexts
	if constexpr (emper::IO) {
		// The global io_uring needs at least workerCount entries in its SQ because
		// for each worker's IoContext one eventfd read is prepared before the
		// globalCompleter is started and submits all previously prepared sqes.
		globalIo = new IoContext(workerCount);
		ioContexts = new IoContext[workerCount];

		if constexpr (emper::STATS) {
			globalIo->stats.workerId = -1;
			std::atexit(emper::io::Stats::printWorkerStats);
		}

		// submit the workers' CQ eventfds to the global IoContext
		for (workerid_t i = 0; i < workerCount; ++i) {
			ioContexts[i].submit_efd();
			if constexpr (emper::STATS) {
				ioContexts[i].stats.workerId = i;
			}
		}

		// start the globalCompleter after all eventfd are submitted so we don't need to
		// synchronize the globalIo's SQ
		IoContext::startGlobalCompleter(*globalIo);
	}

	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_attr_t attr;
		errno = pthread_attr_init(&attr);
		if (errno) DIE;

		// Start non portable.
		cpu_set_t cpuset;
		CPU_ZERO(&cpuset);
		CPU_SET(i % nprocs, &cpuset);

		errno = pthread_attr_setaffinity_np(&attr, sizeof(cpuset), &cpuset);
		if (errno) DIE_MSG_ERRNO("pthread_attr_setaffinity_np() failed");
		// End non portable.

		// This is not done by each thread individually because the randomEngine is
		// mutated and therefore produces data races.
		unsigned int seed = uniformIntDistribution(randomEngine);
		auto* worker = new Worker(i, seed);
		workers[i] = worker;

		auto thread_function = [](void* voidWorkerPtr) -> void* {
#ifdef EMPER_LIBURCU
			rcu_register_thread();
#endif

			auto* worker = reinterpret_cast<Worker*>(voidWorkerPtr);
			return currentRuntime->workerLoop(worker);
		};
		errno = pthread_create(&threads[i], &attr, thread_function, worker);
		if (errno) DIE_MSG_ERRNO("pthread_create() failed");
	}

	if constexpr (emper::STATS) {
		int res = std::atexit(&printLastRuntimeStats);
		if (res) {
			DIE_MSG("could not register printStats() with at_exit()");
		}
	}

	const std::string stacktraceOnAbortsEnvVar("EMPER_STACKTRACE_ON_ABORTS");
	const char* envValueEnableStacktraces = std::getenv(stacktraceOnAbortsEnvVar.c_str());
	if (envValueEnableStacktraces && !std::strcmp(envValueEnableStacktraces, "true")) {
		DBG(stacktraceOnAbortsEnvVar << " set to 'true': Enabling stacktrace on abort");
		enableStacktraceOnAborts();
	}
}

Runtime::~Runtime() {
	DBG("Runtime " << this << " is terminating");
	for (workerid_t i = 0; i < workerCount; ++i) {
		DBG("Runtime " << this << " is cancelling worker " << unsigned(i));
		errno = pthread_cancel(threads[i]);
		if (errno) {
			DIE_MSG_ERRNO("pthread_cancel() failed");
		}
	}
	for (unsigned int i = 0; i < workerCount; ++i) {
		delete workers[i];
	}
	delete[] workers;
	delete[] threads;

	if constexpr (emper::IO) {
		delete[] ioContexts;
		delete globalIo;
	}

	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		currentRuntime = nullptr;
	}
	DBG("Runtime " << this << " terminated");
}

auto Runtime::workerLoop(Worker* worker) -> void* {
	worker->setWorker();

	if constexpr (emper::IO) {
		ioContexts[worker->workerId].setWorkerIo();
	}

	LOGD("Worker loop started by thread " << syscall(SYS_gettid));

	int oldType;
	errno = pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldType);
	if (errno) {
		DIE_MSG_ERRNO("pthread_setcanceltype() failed");
	}

	for (const auto& f : newWorkerHooks) f();

	workerLatch.count_down_and_wait();

	contextManager.start();

	abort();
	return nullptr;
}

auto Runtime::nextFiber() -> NextFiberResult {
	if constexpr (emper::IO) {
		// Schedule all fibers waiting on completed IO
		IoContext::getWorkerIo()->reapCompletions();
	}

	return scheduler.nextFiber();
}

void Runtime::waitUntilFinished() {
	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_join(threads[i], nullptr);
	}
}

void Runtime::printStats() {
	auto runtimeStrategyStats = strategy->getStats();
	runtimeStrategyStats->print();
}

void Runtime::printLastRuntimeStats() {
	std::lock_guard<std::mutex> lock(currentRuntimeMutex);
	if (currentRuntime == nullptr) {
		LOGE("Could not print current runtime stats");
		return;
	}

	currentRuntime->printStats();
}

void Runtime::executeAndWait(std::function<void()> f) {
	if (inRuntime()) {
		ABORT("Ca not use executeAndWait() from within the Runtime");
	}

	std::mutex fiberFinished;
	fiberFinished.lock();

	Fiber* fiber = Fiber::from([&] {
		f();

		fiberFinished.unlock();
	});

	scheduleFromAnywhere(*fiber);

	fiberFinished.lock();
}
