// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "Future.hpp"

#include <sys/types.h>	// for ssize_t

#include <cassert>	// for assert
#include <cerrno>		// for errno, EAGAIN, EWOULDBLOCK
#include <ostream>	// for operator<<, ostream, basic_ost...

#include "BinaryPrivateSemaphore.hpp"	 // for BPS
#include "CallerEnvironment.hpp"			 // for CallerEnvironment
#include "Debug.hpp"									 // for LOGD, LOGW
#include "Runtime.hpp"								 // for Runtime
#include "io/IoContext.hpp"						 // for IoContext
#include "io/Stats.hpp"								 // for Stats, Operation

namespace emper::io {

template <CallerEnvironment callerEnvironment>
auto PartialCompletableFuture::tryComplete(int32_t res)
		-> PartialCompletableFuture::CompletionType {
	// We expect partial completions and made progress
	if (!isDependency() && partialCompletion > DISABLE_PARTIAL_COMPLETION && res > 0) {
		partialCompletion += res;
		// write/send: we have not written the whole buffer yet
		// read: we have not filled the buffer yet
		if ((size_t)partialCompletion < len) {
			IoContext* io = IoContext::getIo<callerEnvironment>();
			io->submit<callerEnvironment>(*this);
			return CompletionType::Resubmission;
		}

		return CompletionType::IncrementalCompletion;
	}

	return CompletionType::Completion;
}

// show the compiler our template incarnations
template PartialCompletableFuture::CompletionType
PartialCompletableFuture::tryComplete<CallerEnvironment::EMPER>(int32_t res);
template PartialCompletableFuture::CompletionType
PartialCompletableFuture::tryComplete<CallerEnvironment::ANYWHERE>(int32_t res);

auto Future::wait() -> int32_t {
	LOGD("Waiting on " << this);

	sem.wait();

	if constexpr (emper::DEBUG) {
		state.retrieved = true;
	}

	return returnValue;
}

auto Future::waitAndSetErrno() -> ssize_t {
	int32_t res = wait();
	// move error codes from io_uring into errno
	if (res < 0) {
		errno = -res;
		return -1;
	}

	return res;
}

void Future::submit() {
	assert(Runtime::inRuntime());
	assert(!isSubmitted());
	IoContext* io = IoContext::getWorkerIo();

	LOGD("submit " << this << " to IoContext " << io);

	io->submit(*this);
}

auto Future::cancel() -> int32_t {
	if (dependency) {
		dependency->cancel();
	}

	state.cancelled = true;

	if (!isSubmitted() || isForgotten()) {
		LOGW("Cancelling unsubmitted or forgotten " << this);
		return -ENOENT;
	}

	if (isCompleted()) {
		return returnValue;
	}

	CancelWrapper cancellation(*this);
	cancellation.submitAndWait();
	return wait();
}

/**
 * Internal hiding functions for Stats::recordCompletion double dispatch
 *
 * The polymorphism functions are Future::recordCompletion and are defined in
 * Future.hpp so they are included in each compilation unit and the
 * compiler can optimize the recordCompletion calls if stats are off
 */
void Future::recordCompletionInternal(Stats& stats, int32_t res) const {
	stats.recordCompletion(*this, res);
}

void PartialCompletableFuture::recordCompletionInternal(Stats& stats, int32_t res) const {
	stats.recordCompletion(*this, res);
}

void WritevFuture::recordCompletionInternal(Stats& stats, int32_t res) const {
	stats.recordCompletion(*this, res);
}
}	 // namespace emper::io
