// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#pragma once

#include <liburing.h>	 // for io_uring
#include <pthread.h>	 // for pthread_t

#include <atomic>			 // for atomic
#include <cassert>		 // for assert
#include <cstddef>		 // for size_t
#include <cstdint>		 // for uint64_t
#include <functional>	 // for less
#include <memory>			 // for allocator
#include <mutex>			 // for mutex

#include "CallerEnvironment.hpp"	// for CallerEnvironment, EMPER
#include "Debug.hpp"							// for LogSubsystem, LogSubsystem::IO, Logger
#include "Runtime.hpp"						// for Runtime
#include "emper-config.h"					// for EMPER_IO_WORKER_URING_ENTRIES
#include "io/Stats.hpp"						// for Stats
#include "lib/adt/LockedSet.hpp"	// for LockedSet

namespace emper::io {
class Future;

class IoContext : public Logger<LogSubsystem::IO> {
	friend class ::Runtime;
	friend class Future;
	friend class SendFuture;
	friend class RecvFuture;

 private:
	static thread_local IoContext *workerIo;

	// pthread used to monitor the CQs from worker io_urings
	// as well as handling IO requests submitted from anywhere
	static pthread_t globalCompleter;

	/* function executed by the global completer thread */
	static auto globalCompleterFunc(void *arg) -> void *;

	// start the global completer thread
	// this must be called after all worker IoContexts' eventfds are submitted
	// to the global IoContext
	static void startGlobalCompleter(IoContext &globalIo);

	// Mutex protecting the completion queue of ring.
	// It is used with try_lock() in reapCompletions.
	std::mutex cq_mutex;
	struct io_uring ring;

	// eventfd registered with ring.
	// It is used to notify the globalCompleter about available completions in ring's CQ.
	int ring_eventfd;
	uint64_t ring_eventfd_readbuf;

	// submit the eventfd of the io_uring to the global IoContext
	// All eventfds must be registered before the globalCompleter thread is started
	void submit_efd();

	Stats stats;

	// Members useful for debugging
	std::atomic<long> reqs_in_uring = 0;

	// set containing all Futures currently in the io_uring for debugging
	emper::lib::adt::LockedSet<Future *> uringFutureSet;

	/**
	 * @brief recursively prepare a Future and its dependencies
	 *
	 * Futures prepared as dependency will have disabled partial completion
	 * support and are submitted to io_uring with the IOSQE_IO_LINK flag
	 * set.
	 * Abort on preparing a longer list then there a places in the io_uring.
	 *
	 * @return the number of prepared Futures
	 */
	auto prepareFutureChain(Future &future, unsigned chain_length) -> unsigned;

	inline void setWorkerIo() { workerIo = this; }

	/**
	 * @brief Return the globalIo of the current Runtime
	 */
	static inline auto getGlobalIo() -> IoContext * { return Runtime::getRuntime()->getGlobalIo(); }

 public:
	IoContext(size_t uring_entries);
	IoContext() : IoContext(EMPER_IO_WORKER_URING_ENTRIES){};
	~IoContext();

	/**
	 * @brief get IoContext of the current worker
	 *
	 * This function must be called from with in the emper runtime
	 *
	 * @return the thread_local IoContext of the current worker
	 */
	static inline auto getWorkerIo() -> IoContext * {
		assert(Runtime::inRuntime());
		return workerIo;
	}

	/**
	 * @brief Return either the worker or the globalIo dependent on the callerEnvironment
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	static inline auto getIo() -> IoContext * {
		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			return getWorkerIo();
		} else {
			return getGlobalIo();
		}
	}

	/**
	 * @brief Submit a future for asynchronous completion in this IoContext
	 *
	 * @param future The Future which should be completed. The Future object
	 *        must stay valid until it is completed.
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void submit(Future &future);

	/**
	 * @brief Schedule all fibers waiting on completed IO
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void reapCompletions();
};
}	 // namespace emper::io
