// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Fischer
#include "io/Stats.hpp"

#include <cassert>					 // for assert
#include <initializer_list>	 // for initializer_list
#include <iomanip>					 // for operator<<, setfill, setw
#include <string>						 // for operator<<, to_string
#include <utility>					 // for pair

#include "io/Operation.hpp"	 // for Operation, operator<<, Operation::RECV

namespace emper::io {

std::vector<Stats*> Stats::workerStats;

auto operator<<(std::ostream& os, const Stats::CompletionType& t) -> std::ostream& {
	switch (t) {
		case Stats::CompletionType::FullCompletion:
			os << "full completion";
			break;
		case Stats::CompletionType::ErrorCompletion:
			os << "error completion";
			break;
		case Stats::CompletionType::PartialCompletion:
			os << "partial completion";
			break;
		case Stats::CompletionType::PartialResubmission:
			os << "partial resubmission";
			break;
		case Stats::CompletionType::IncrementalCompletion:
			os << "incremental completion";
			break;
		case Stats::CompletionType::IncrementalError:
			os << "incremental error";
			break;
		default:
			abort();
	}
	return os;
}

auto operator<<(std::ostream& os, const Stats& s) -> std::ostream& {
	if (s.workerId == -1) {
		bool futuresToPrint = false;
		std::stringstream ss;
		ss << "# global IO Stats: #" << std::endl;
		// Only PartialCompletableFutures can submitted to the globalIo
		for (const auto& op : {Operation::SEND, Operation::READ, Operation::WRITE}) {
			const auto& op_map = s.io_uring_completions.at(op);

			ss << op << ": ";
			ss << std::left << std::setfill(' ') << std::setw(8) << ss.str();

			for (const auto& tp : op_map) {
				auto completionType = tp.first;
				auto count = tp.second;

				if (count > 0) {
					futuresToPrint = true;
				}

				ss << count << " " << completionType << " ";
			}
			ss << std::endl;
		}

		if (futuresToPrint) {
			os << ss.str();
		}

		return os;
	}

	if (s.workerId == -2) {
		os << "# AVG IO worker Stats: #" << std::endl;
	} else {
		os << "# IO Stats of worker " << std::to_string(s.workerId) << ": #" << std::endl;
	}

	for (const auto& pp : s.io_uring_completions) {
		const auto& op = pp.first;
		const auto& op_map = pp.second;

		std::stringstream ss;
		ss << op << ": ";
		os << std::left << std::setfill(' ') << std::setw(16) << ss.str();

		for (const auto& tp : op_map) {
			auto completionType = tp.first;
			auto count = tp.second;
			os << count << " " << completionType << " ";
		}

		os << std::endl;
	}

	if (s.workerId == -1) {
		return os;
	}

	os << "io_uring_submit reap completion loops: " << s.io_submit_full_cq << " taking "
		 << s.io_submit_full_cq_running_mean << "ns on average" << std::endl;

	os << "worker reaped completions " << s.worker_reap << " times reaping "
		 << s.worker_reaped_completions << " completions in total" << std::endl;

	os << "global completer thread reaped completions " << s.completer_reap << " times reaping "
		 << s.completer_reaped_completions << " completions in total" << std::endl;

	return os;
}

static void calcRunningAvg(uint64_t& avg, int n, uint64_t x) {
	int64_t diff = ((int64_t)x - (int64_t)avg) / n;
	assert(diff >= 0 || (uint64_t)-diff <= avg);
	avg += diff;
	assert(avg < 1000000);
}

void Stats::printWorkerStats() {
	auto* globalStats = workerStats.front();
	workerStats.erase(workerStats.begin());
	std::cout << *globalStats << std::endl;

	// Use a stats object to calculate the averages
	Stats avgs;
	// mark as averages
	avgs.workerId = -2;

	// calculate avgs
	int i = 1;
	for (auto& stats : workerStats) {
		for (const auto& op_pair : stats->io_uring_completions) {
			auto op = op_pair.first;
			auto op_map = op_pair.second;

			for (const auto& compl_pair : op_map) {
				auto compl_typ = compl_pair.first;
				auto compl_count = compl_pair.second;
				calcRunningAvg(avgs.io_uring_completions[op][compl_typ], i, compl_count);
			}
		}

		calcRunningAvg(avgs.io_submit_full_cq, i, stats->io_submit_full_cq);
		calcRunningAvg(avgs.io_submit_full_cq_running_mean, i, stats->io_submit_full_cq_running_mean);

		calcRunningAvg(avgs.completer_reap, i, stats->completer_reap);
		calcRunningAvg(avgs.completer_reaped_completions, i, stats->completer_reaped_completions);

		calcRunningAvg(avgs.worker_reap, i, stats->worker_reap);
		calcRunningAvg(avgs.worker_reaped_completions, i, stats->worker_reaped_completions);
		++i;
	}

	// print avgs
	std::cout << avgs << std::endl;
}

}	 // namespace emper::io
