// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Schmaus
#include "AbstractWorkStealingScheduler.hpp"

#include <atomic>
#include <ostream>	// for operator<<, basic_ostream<>::__ostream_type

#include "Common.hpp"	 // for unlikely, likely
#include "Debug.hpp"	 // for ABORT
#include "Emper.hpp"	 // for OVERFLOW_QUEUE
#include "NextFiberResult.hpp"
#include "Runtime.hpp"		 // for Runtime
#include "emper-common.h"	 // for workerid_t
#include "strategies/AbstractWorkStealingStrategy.hpp"

class Fiber;

thread_local AbstractWorkStealingScheduler::WsQueue<AbstractWorkStealingScheduler::QUEUE_SIZE>
		AbstractWorkStealingScheduler::queue;

AbstractWorkStealingScheduler::AbstractWorkStealingScheduler(
		Runtime& runtime, AbstractWorkStealingStrategy& abstractWorkStealingStrategy)
		: Scheduler(runtime), abstractWorkStealingStrategy(abstractWorkStealingStrategy) {
	const workerid_t workerCount = runtime.getWorkerCount();
	queues = new AbstractWorkStealingScheduler::WsQueue<QUEUE_SIZE>*[workerCount];

	auto newWorkerHook = [this]() { queues[Runtime::getWorkerId()] = &queue; };
	addNewWorkerHook(newWorkerHook);
}

void AbstractWorkStealingScheduler::scheduleViaWorkStealing(Fiber& fiber) {
	bool pushed = queue.pushBottom(&fiber);
	if (unlikely(!pushed)) {
		if constexpr (emper::OVERFLOW_QUEUE) {
			enqueueInAnywhereQueue(fiber);

			if constexpr (emper::STATS) {
				// TODO: Use template magic so that this becomes
				// incrementRelaxed(abstractWorkStealingStrategy.scheduledFibersToLocal)
				abstractWorkStealingStrategy.scheduledFibersToOverflowQueue.fetch_add(
						1, std::memory_order_relaxed);
			}
		} else {
			ABORT("Could not push fiber " << &fiber << " into queue");
		}
	} else if constexpr (emper::STATS) {
		abstractWorkStealingStrategy.scheduledFibersToLocal.fetch_add(1, std::memory_order_relaxed);
	}

	// Classes using this method are supposed to always invoke this
	// method. Hence we call onNewWork() here.
	onNewWork();
}

auto AbstractWorkStealingScheduler::nextFiberViaWorkStealing() -> std::pair<Fiber*, FiberSource> {
	FiberSource fiberSource = FiberSource::local;
	Fiber* fiber;

	bool poped = queue.popBottom(&fiber);
	if (likely(poped)) {
		if constexpr (emper::STATS) {
			abstractWorkStealingStrategy.nextFiberFromLocal.fetch_add(1, std::memory_order_relaxed);
		}

		goto out;
	}

	{
		const workerid_t myWorkerId = Runtime::getWorkerId();
		const workerid_t workerCount = runtime.getWorkerCount();
		workerid_t startWorkerId = Runtime::rand() % workerCount;
		// TODO: See how reducing the loop bound affects things.
		for (workerid_t i = 0; i < workerCount; ++i) {
			workerid_t victim = (startWorkerId + i) % workerCount;

			// Don't steal from ourselves.
			if (unlikely(victim == myWorkerId)) continue;

			poped = queues[victim]->popTop(&fiber);
			if (poped) {
				if constexpr (emper::STATS) {
					abstractWorkStealingStrategy.nextFiberStolen.fetch_add(1, std::memory_order_relaxed);
				}

				fiberSource = FiberSource::stolen;
				goto out;
			}
		}
	}

	// Try the "scheduled from anywhere" queue to get work as last resort.
	fiber = dequeueFiberFromAnywhereQueue();
	if (fiber) {
		if constexpr (emper::STATS) {
			abstractWorkStealingStrategy.nextFiberFromAnywhereQueue.fetch_add(1,
																																				std::memory_order_relaxed);
		}

		fiberSource = FiberSource::anywhereQueue;
		goto out;
	}

out:
	return std::make_pair(fiber, fiberSource);
}

auto AbstractWorkStealingScheduler::nextFiberResultViaWorkStealing() -> NextFiberResult {
	std::pair<Fiber*, FiberSource> nextFiberWsResult = nextFiberViaWorkStealing();
	return NextFiberResult{
			nextFiberWsResult.first,
			static_cast<uintptr_t>(nextFiberWsResult.second),
	};
}
