// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus
#include "LawsScheduler.hpp"

#include <atomic>	 // for atomic, memory_order_relaxed
#include <cstdint>

#include "Emper.hpp"
#include "LawsStrategy.hpp"	 // IWYU pragma: keep
#include "NextFiberResult.hpp"
#include "Runtime.hpp"

#define EMPER_OVERFLOW_QUEUE

thread_local LawsScheduler::LawsMpscQueue LawsScheduler::priorityQueue;

LawsScheduler::LawsScheduler(Runtime& runtime, LawsStrategy& lawsStrategy)
		: AbstractWorkStealingScheduler(runtime, lawsStrategy), lawsStrategy(lawsStrategy) {
	const workerid_t workerCount = runtime.getWorkerCount();
	priorityQueues = new LawsScheduler::LawsMpscQueue*[workerCount];

	auto newWorkerHook = [this]() {
		workerid_t workerId = Runtime::getWorkerId();
		priorityQueues[workerId] = &priorityQueue;
	};
	addNewWorkerHook(newWorkerHook);
}

void LawsScheduler::scheduleInternal(Fiber& fiber) {
	workeraffinity_t* const affinity_buffer = getAffinityBuffer(fiber);
	if (affinity_buffer) {
		workeraffinity_t affinity = *affinity_buffer;
		workerid_t workerId = Runtime::getWorkerId();
		if (affinity == workerId) {
			goto scheduleViaWorkStealing;
		}

		if (affinity == Fiber::NOT_AFFINE) {
			goto scheduleViaWorkStealing;
		}

		// We found a fiber to schedule on a remote prority queue.
		increaseRefCount(fiber);
		priorityQueues[affinity]->enqueue(&fiber);
		if constexpr (emper::STATS) {
			lawsStrategy.scheduledFibersToRemotePriority.fetch_add(1, std::memory_order_relaxed);
		}
	}

scheduleViaWorkStealing:
	scheduleViaWorkStealing(fiber);
}

auto LawsScheduler::nextFiber() -> NextFiberResult {
	Fiber* fiber = priorityQueue.dequeue();
	if (fiber != nullptr) {
		return NextFiberResult{
				fiber,
				static_cast<uintptr_t>(LawsStrategy::FiberSource::fromPriority),
		};
	}

	return nextFiberResultViaWorkStealing();
}
